#ifndef _MOVIT_BUNDLED_SHADERS_H
#define _MOVIT_BUNDLED_SHADERS_H 1

// Shader files that are compiled into the library.

namespace movit {

struct BundledShader {
	const char *filename;  // nullptr for end.
	size_t offset;  // Into shader_bundle.
	size_t length;
};
extern BundledShader bundled_shaders[];
extern const char *shader_bundle;

}  // namespace movit

#endif  // _MOVIT_BUNDLED_SHADERS_H
